from typing import Any
from typing import Union

from scalecodec import ScaleType

from libtensorshield.utils import decode_account_id
from ._fixedpoint import FixedPoint


class Balance:
    """
    Represents the bittensor balance of the wallet, stored as rao (int).
    This class provides a way to interact with balances in two different units: rao and tao.
    It provides methods to convert between these units, as well as to perform arithmetic and comparison operations.

    Attributes:
        unit (str): A string representing the symbol for the tao unit.
        rao_unit (str): A string representing the symbol for the rao unit.
        rao (int): An integer that stores the balance in rao units.
        tao (float): A float property that gives the balance in tao units.
    """
    TAO_SYMBOL: str = chr(0x03C4)
    RAO_SYMBOL: str = chr(0x03C1)

    unit: str = TAO_SYMBOL  # This is the tao unit
    rao_unit: str = RAO_SYMBOL  # This is the rao unit
    rao: int

    def __init__(self, balance: Union[int, float, 'Balance']):
        """
        Initialize a Balance object. If balance is an int, it's assumed to be in rao.
        If balance is a float, it's assumed to be in tao.

        Args:
            balance: The initial balance, in either rao (if an int) or tao (if a float).
        """
        if not isinstance(balance, (int, float, Balance)): # type: ignore
            raise TypeError("balance must be an int (rao) or a float (tao)") # pragma: no cover
        if isinstance(balance, int):
            self.rao = balance
        elif isinstance(balance, Balance):
            self.rao = balance.rao
        else:
            assert isinstance(balance, float)
            # Assume tao value for the float
            self.rao = int(balance * pow(10, 9))

    @property
    def tao(self) -> float:
        return self.rao / pow(10, 9)

    @staticmethod
    def process_stake_data(stake_data: list[tuple[bytes, int]]) -> dict[str, 'Balance']:
        """
        Processes stake data to decode account IDs and convert stakes from rao to Balance objects.

        Args:
            stake_data (list): A list of tuples where each tuple contains an account ID in bytes and a stake in rao.

        Returns:
            dict: A dictionary with account IDs as keys and their corresponding Balance objects as values.
        """
        decoded_stake_data: dict[str, Balance] = {}
        for account_id_bytes, stake_ in stake_data:
            account_id = decode_account_id(account_id_bytes)
            decoded_stake_data.update({account_id: Balance.from_rao(stake_)})
        return decoded_stake_data

    def __int__(self):
        """Convert the Balance object to an int. The resulting value is in rao."""
        return self.rao

    def __float__(self):
        """Convert the Balance object to a float. The resulting value is in tao."""
        return float(self.tao)

    def __str__(self): # pragma: no cover
        """
        Returns the Balance object as a string in the format "symbolvalue", where the value is in tao.
        """
        if self.unit == units[0]:
            return f"{self.unit}{float(self.tao):,.9f}"
        else:
            return f"\u200e{float(self.tao):,.9f}{self.unit}\u200e"

    def __rich__(self): # pragma: no cover
        int_tao, fract_tao = format(float(self.tao), "f").split(".")
        return f"[green]{self.unit}{int_tao}.{fract_tao}[/green]"

    def __str_rao__(self):  # pragma: no cover
        return f"{self.rao_unit}{int(self.rao)}"

    def __rich_rao__(self):  # pragma: no cover
        return f"[green]{self.rao_unit}{int(self.rao)}[/green]"

    def __repr__(self):  # pragma: no cover
        return self.__str__()

    def __eq__(self, other: Any):
        if other is None:
            return False
        if not isinstance(other, (int, float, Balance)): # pragma: no cover
            return NotImplemented

        if isinstance(other, Balance):
            return self.rao == other.rao
        else:
            return self.rao == int(other)

    def __ne__(self, other: Any):
        if not isinstance(other, (int, float, Balance)): # pragma: no cover
            return NotImplemented
        return not self == other

    def __gt__(self, other: Any):
        if not isinstance(other, (int, float, Balance)): # pragma: no cover
            return NotImplemented
        if isinstance(other, Balance):
            return self.rao > other.rao
        else:
            return self.rao > int(other)

    def __lt__(self, other: Any):
        if not isinstance(other, (int, float, type(self))): # pragma: no cover
            return NotImplemented
        match isinstance(other, type(self)):
            case True:
                assert isinstance(other, type(self))
                return self.rao < other.rao
            case False:
                return self.rao < int(other)

    def __le__(self, other: Any): # pragma: no cover
        if not isinstance(other, (int, float, Balance)): # pragma: no cover
            return NotImplemented
        return self < other or self == other

    def __ge__(self, other: Any):
        if not isinstance(other, (int, float, Balance)): # pragma: no cover
            return NotImplemented
        return self > other or self == other

    def __add__(self, other: Any):
        if not isinstance(other, (int, float, type(self))): # pragma: no cover
            return NotImplemented
        if isinstance(other, Balance):
            return Balance.from_rao(int(self.rao + other.rao))
        else:
            return Balance.from_rao(int(self.rao + other))

    def __radd__(self, other: Any):
        if not isinstance(other, (int, float, type(self))): # pragma: no cover
            return NotImplemented
        return self + other

    def __sub__(self, other: Any):
        if not isinstance(other, (int, float, type(self))): # pragma: no cover
            return NotImplemented
        return self + -other

    def __rsub__(self, other: Any):
        if not isinstance(other, (int, float, type(self))): # pragma: no cover
            return NotImplemented
        return -self + other

    def __mul__(self, other: Union[int, float, "Balance"]):
        if not isinstance(other, (int, float, type(self))): # pragma: no cover
            return NotImplemented
        if isinstance(other, Balance):
            return Balance.from_rao(int(self.rao * other.rao))
        else:
            return Balance.from_rao(int(self.rao * other))

    def __rmul__(self, other: Any):
        return self * other

    def __truediv__(self, other: Any):
        if not isinstance(other, (int, float, type(self))): # pragma: no cover
            return NotImplemented
        if isinstance(other, Balance):
            return Balance.from_rao(int(self.rao / other.rao))
        else:
            return Balance.from_rao(int(self.rao / other))

    def __rtruediv__(self, other: Any):
        if not isinstance(other, (int, float, type(self))): # pragma: no cover
            return NotImplemented
        if isinstance(other, Balance): # pragma: no cover
            return Balance.from_rao(int(other.rao / self.rao))
        else:
            return Balance.from_rao(int(other / self.rao))

    #def __floordiv__(self, other: Any):
    #    if not isinstance(other, (int, float, type(self))): # pragma: no cover
    #        return NotImplemented
    #    if isinstance(other, Balance):
    #        return Balance.from_rao(int(self.tao // other.tao))
    #    else:
    #        return Balance.from_rao(int(self.rao // other))

    #def __rfloordiv__(self, other: Any):
    #    if not isinstance(other, (int, float, Balance)): # pragma: no cover
    #        return NotImplemented
    #    if isinstance(other, Balance):
    #        return Balance.from_rao(int(other.rao // self.rao))
    #    else:
    #        return Balance.from_rao(int(other // self.rao))

    def __nonzero__(self) -> bool: # pragma: no cover
        return bool(self.rao)

    def __neg__(self):  # pragma: no cover
        return Balance.from_rao(-self.rao)

    def __pos__(self):  # pragma: no cover
        return Balance.from_rao(self.rao)

    def __abs__(self):
        return Balance.from_rao(abs(self.rao))

    @staticmethod
    def from_float(amount: float, netuid: int = 0) -> 'Balance':
        """
        Given tao, return :func:`Balance` object with rao(``int``) and tao(``float``), where rao = int(tao*pow(10,9))
        Args:
            amount (float): The amount in tao.
            netuid (int): The subnet uid for set currency unit. Defaults to `0`.

        Returns:
            A Balance object representing the given amount.
        """
        rao_ = int(amount * pow(10, 9))
        return Balance(rao_).set_unit(netuid)

    @staticmethod
    def from_tao(amount: float, netuid: int = 0):
        """
        Given tao, return Balance object with rao(``int``) and tao(``float``), where rao = int(tao*pow(10,9))

        Args:
            amount (float): The amount in tao.
            netuid (int): The subnet uid for set currency unit. Defaults to `0`.

        Returns:
            A Balance object representing the given amount.
        """
        rao_ = int(amount * pow(10, 9))
        return Balance(rao_).set_unit(netuid)

    @staticmethod
    def from_rao(amount: int, netuid: int = 0):
        """
        Given rao, return Balance object with rao(``int``) and tao(``float``), where rao = int(tao*pow(10,9))

        Args:
            amount (int): The amount in rao.
            netuid (int): The subnet uid for set currency unit. Defaults to `0`.

        Returns:
            A Balance object representing the given amount.
        """
        return Balance(amount).set_unit(netuid)

    @staticmethod
    def get_unit(netuid: int): # pragma: no cover
        base = len(units)
        if netuid < base:
            return units[netuid]
        else:
            result = ""
            while netuid > 0:
                result = units[netuid % base] + result
                netuid //= base
            return result

    def set_unit(self, netuid: int):
        self.unit = Balance.get_unit(netuid)
        self.rao_unit = Balance.get_unit(netuid)
        return self

    @staticmethod
    def fixed_to_float(
        fixed: FixedPoint | ScaleType,
        frac_bits: int = 64,
        total_bits: int = 128
    ) -> float:  # pragma: no cover
        # By default, this is a U64F64
        # which is 64 bits of integer and 64 bits of fractional

        data: int = fixed["bits"] # type: ignore
        assert isinstance(data, int)

        # Logical and to get the fractional part; remaining is the integer part
        fractional_part = data & (2**frac_bits - 1)
        # Shift to get the integer part from the remaining bits
        integer_part = data >> (total_bits - frac_bits)

        frac_float = fractional_part / (2**frac_bits)

        return integer_part + frac_float


units = [
    # Greek Alphabet (0-24)
    "\u03c4",  # τ (tau, 0)
    "\u03b1",  # α (alpha, 1)
    "\u03b2",  # β (beta, 2)
    "\u03b3",  # γ (gamma, 3)
    "\u03b4",  # δ (delta, 4)
    "\u03b5",  # ε (epsilon, 5)
    "\u03b6",  # ζ (zeta, 6)
    "\u03b7",  # η (eta, 7)
    "\u03b8",  # θ (theta, 8)
    "\u03b9",  # ι (iota, 9)
    "\u03ba",  # κ (kappa, 10)
    "\u03bb",  # λ (lambda, 11)
    "\u03bc",  # μ (mu, 12)
    "\u03bd",  # ν (nu, 13)
    "\u03be",  # ξ (xi, 14)
    "\u03bf",  # ο (omicron, 15)
    "\u03c0",  # π (pi, 16)
    "\u03c1",  # ρ (rho, 17)
    "\u03c3",  # σ (sigma, 18)
    "t",  # t (tau, 19)
    "\u03c5",  # υ (upsilon, 20)
    "\u03c6",  # φ (phi, 21)
    "\u03c7",  # χ (chi, 22)
    "\u03c8",  # ψ (psi, 23)
    "\u03c9",  # ω (omega, 24)
    # Hebrew Alphabet (25-51)
    "\u05d0",  # א (aleph, 25)
    "\u05d1",  # ב (bet, 26)
    "\u05d2",  # ג (gimel, 27)
    "\u05d3",  # ד (dalet, 28)
    "\u05d4",  # ה (he, 29)
    "\u05d5",  # ו (vav, 30)
    "\u05d6",  # ז (zayin, 31)
    "\u05d7",  # ח (het, 32)
    "\u05d8",  # ט (tet, 33)
    "\u05d9",  # י (yod, 34)
    "\u05da",  # ך (final kaf, 35)
    "\u05db",  # כ (kaf, 36)
    "\u05dc",  # ל (lamed, 37)
    "\u05dd",  # ם (final mem, 38)
    "\u05de",  # מ (mem, 39)
    "\u05df",  # ן (final nun, 40)
    "\u05e0",  # נ (nun, 41)
    "\u05e1",  # ס (samekh, 42)
    "\u05e2",  # ע (ayin, 43)
    "\u05e3",  # ף (final pe, 44)
    "\u05e4",  # פ (pe, 45)
    "\u05e5",  # ץ (final tsadi, 46)
    "\u05e6",  # צ (tsadi, 47)
    "\u05e7",  # ק (qof, 48)
    "\u05e8",  # ר (resh, 49)
    "\u05e9",  # ש (shin, 50)
    "\u05ea",  # ת (tav, 51)
    # Arabic Alphabet (52-81)
    "\u0627",  # ا (alif, 52)
    "\u0628",  # ب (ba, 53)
    "\u062a",  # ت (ta, 54)
    "\u062b",  # ث (tha, 55)
    "\u062c",  # ج (jeem, 56)
    "\u062d",  # ح (ha, 57)
    "\u062e",  # خ (kha, 58)
    "\u062f",  # د (dal, 59)
    "\u0630",  # ذ (dhal, 60)
    "\u0631",  # ر (ra, 61)
    "\u0632",  # ز (zay, 62)
    "\u0633",  # س (seen, 63)
    "\u0634",  # ش (sheen, 64)
    "\u0635",  # ص (sad, 65)
    "\u0636",  # ض (dad, 66)
    "\u0637",  # ط (ta, 67)
    "\u0638",  # ظ (dha, 68)
    "\u0639",  # ع (ain, 69)
    "\u063a",  # غ (ghain, 70)
    "\u0641",  # ف (fa, 71)
    "\u0642",  # ق (qaf, 72)
    "\u0643",  # ك (kaf, 73)
    "\u0644",  # ل (lam, 74)
    "\u0645",  # م (meem, 75)
    "\u0646",  # ن (noon, 76)
    "\u0647",  # ه (ha, 77)
    "\u0648",  # و (waw, 78)
    "\u064a",  # ي (ya, 79)
    "\u0649",  # ى (alef maksura, 80)
    "\u064a",  # ي (ya, 81)
    # Runic Alphabet (82-90)
    "\u16a0",  # ᚠ (fehu, 82)
    "\u16a2",  # ᚢ (uruz, 83)
    "\u16a6",  # ᚦ (thurisaz, 84)
    "\u16a8",  # ᚨ (ansuz, 85)
    "\u16b1",  # ᚱ (raidho, 86)
    "\u16b3",  # ᚲ (kaunan, 87)
    "\u16c7",  # ᛇ (eihwaz, 88)
    "\u16c9",  # ᛉ (algiz, 89)
    "\u16d2",  # ᛒ (berkanan, 90)
    # Ogham Alphabet (91-97)
    "\u1680",  #   (Space, 91)
    "\u1681",  # ᚁ (Beith, 92)
    "\u1682",  # ᚂ (Luis, 93)
    "\u1683",  # ᚃ (Fearn, 94)
    "\u1684",  # ᚄ (Sail, 95)
    "\u1685",  # ᚅ (Nion, 96)
    "\u169b",  # ᚛ (Forfeda, 97)
    # Georgian Alphabet (98-103)
    "\u10d0",  # ა (ani, 98)
    "\u10d1",  # ბ (bani, 99)
    "\u10d2",  # გ (gani, 100)
    "\u10d3",  # დ (doni, 101)
    "\u10d4",  # ე (eni, 102)
    "\u10d5",  # ვ (vini, 103)
    # Armenian Alphabet (104-110)
    "\u0531",  # Ա (Ayp, 104)
    "\u0532",  # Բ (Ben, 105)
    "\u0533",  # Գ (Gim, 106)
    "\u0534",  # Դ (Da, 107)
    "\u0535",  # Ե (Ech, 108)
    "\u0536",  # Զ (Za, 109)
    "\u055e",  # ՞ (Question mark, 110)
    # Cyrillic Alphabet (111-116)
    "\u0400",  # Ѐ (Ie with grave, 111)
    "\u0401",  # Ё (Io, 112)
    "\u0402",  # Ђ (Dje, 113)
    "\u0403",  # Ѓ (Gje, 114)
    "\u0404",  # Є (Ukrainian Ie, 115)
    "\u0405",  # Ѕ (Dze, 116)
    # Coptic Alphabet (117-122)
    "\u2c80",  # Ⲁ (Alfa, 117)
    "\u2c81",  # ⲁ (Small Alfa, 118)
    "\u2c82",  # Ⲃ (Vida, 119)
    "\u2c83",  # ⲃ (Small Vida, 120)
    "\u2c84",  # Ⲅ (Gamma, 121)
    "\u2c85",  # ⲅ (Small Gamma, 122)
    # Brahmi Script (123-127)
    "\U00011000",  # 𑀀 (A, 123)
    "\U00011001",  # 𑀁 (Aa, 124)
    "\U00011002",  # 𑀂 (I, 125)
    "\U00011003",  # 𑀃 (Ii, 126)
    "\U00011005",  # 𑀅 (U, 127)
    # Tifinagh Alphabet (128-133)
    "\u2d30",  # ⴰ (Ya, 128)
    "\u2d31",  # ⴱ (Yab, 129)
    "\u2d32",  # ⴲ (Yabh, 130)
    "\u2d33",  # ⴳ (Yag, 131)
    "\u2d34",  # ⴴ (Yagh, 132)
    "\u2d35",  # ⴵ (Yaj, 133)
    # Glagolitic Alphabet (134-166)
    "\u2c00",  # Ⰰ (Az, 134)
    "\u2c01",  # Ⰱ (Buky, 135)
    "\u2c02",  # Ⰲ (Vede, 136)
    "\u2c03",  # Ⰳ (Glagoli, 137)
    "\u2c04",  # Ⰴ (Dobro, 138)
    "\u2c05",  # Ⰵ (Yest, 139)
    "\u2c06",  # Ⰶ (Zhivete, 140)
    "\u2c07",  # Ⰷ (Zemlja, 141)
    "\u2c08",  # Ⰸ (Izhe, 142)
    "\u2c09",  # Ⰹ (Initial Izhe, 143)
    "\u2c0a",  # Ⰺ (I, 144)
    "\u2c0b",  # Ⰻ (Djerv, 145)
    "\u2c0c",  # Ⰼ (Kako, 146)
    "\u2c0d",  # Ⰽ (Ljudije, 147)
    "\u2c0e",  # Ⰾ (Myse, 148)
    "\u2c0f",  # Ⰿ (Nash, 149)
    "\u2c10",  # Ⱀ (On, 150)
    "\u2c11",  # Ⱁ (Pokoj, 151)
    "\u2c12",  # Ⱂ (Rtsy, 152)
    "\u2c13",  # Ⱃ (Slovo, 153)
    "\u2c14",  # Ⱄ (Tvrido, 154)
    "\u2c15",  # Ⱅ (Uku, 155)
    "\u2c16",  # Ⱆ (Fert, 156)
    "\u2c17",  # Ⱇ (Xrivi, 157)
    "\u2c18",  # Ⱈ (Ot, 158)
    "\u2c19",  # Ⱉ (Cy, 159)
    "\u2c1a",  # Ⱊ (Shcha, 160)
    "\u2c1b",  # Ⱋ (Er, 161)
    "\u2c1c",  # Ⱌ (Yeru, 162)
    "\u2c1d",  # Ⱍ (Small Yer, 163)
    "\u2c1e",  # Ⱎ (Yo, 164)
    "\u2c1f",  # Ⱏ (Yu, 165)
    "\u2c20",  # Ⱐ (Ja, 166)
    # Thai Alphabet (167-210)
    "\u0e01",  # ก (Ko Kai, 167)
    "\u0e02",  # ข (Kho Khai, 168)
    "\u0e03",  # ฃ (Kho Khuat, 169)
    "\u0e04",  # ค (Kho Khon, 170)
    "\u0e05",  # ฅ (Kho Rakhang, 171)
    "\u0e06",  # ฆ (Kho Khwai, 172)
    "\u0e07",  # ง (Ngo Ngu, 173)
    "\u0e08",  # จ (Cho Chan, 174)
    "\u0e09",  # ฉ (Cho Ching, 175)
    "\u0e0a",  # ช (Cho Chang, 176)
    "\u0e0b",  # ซ (So So, 177)
    "\u0e0c",  # ฌ (Cho Choe, 178)
    "\u0e0d",  # ญ (Yo Ying, 179)
    "\u0e0e",  # ฎ (Do Chada, 180)
    "\u0e0f",  # ฏ (To Patak, 181)
    "\u0e10",  # ฐ (Tho Than, 182)
    "\u0e11",  # ฑ (Tho Nangmontho, 183)
    "\u0e12",  # ฒ (Tho Phuthao, 184)
    "\u0e13",  # ณ (No Nen, 185)
    "\u0e14",  # ด (Do Dek, 186)
    "\u0e15",  # ต (To Tao, 187)
    "\u0e16",  # ถ (Tho Thung, 188)
    "\u0e17",  # ท (Tho Thahan, 189)
    "\u0e18",  # ธ (Tho Thong, 190)
    "\u0e19",  # น (No Nu, 191)
    "\u0e1a",  # บ (Bo Baimai, 192)
    "\u0e1b",  # ป (Po Pla, 193)
    "\u0e1c",  # ผ (Pho Phung, 194)
    "\u0e1d",  # ฝ (Fo Fa, 195)
    "\u0e1e",  # พ (Pho Phan, 196)
    "\u0e1f",  # ฟ (Fo Fan, 197)
    "\u0e20",  # ภ (Pho Samphao, 198)
    "\u0e21",  # ม (Mo Ma, 199)
    "\u0e22",  # ย (Yo Yak, 200)
    "\u0e23",  # ร (Ro Rua, 201)
    "\u0e25",  # ล (Lo Ling, 202)
    "\u0e27",  # ว (Wo Waen, 203)
    "\u0e28",  # ศ (So Sala, 204)
    "\u0e29",  # ษ (So Rusi, 205)
    "\u0e2a",  # ส (So Sua, 206)
    "\u0e2b",  # ห (Ho Hip, 207)
    "\u0e2c",  # ฬ (Lo Chula, 208)
    "\u0e2d",  # อ (O Ang, 209)
    "\u0e2e",  # ฮ (Ho Nokhuk, 210)
    # Hangul Consonants (211-224)
    "\u1100",  # ㄱ (Giyeok, 211)
    "\u1101",  # ㄴ (Nieun, 212)
    "\u1102",  # ㄷ (Digeut, 213)
    "\u1103",  # ㄹ (Rieul, 214)
    "\u1104",  # ㅁ (Mieum, 215)
    "\u1105",  # ㅂ (Bieup, 216)
    "\u1106",  # ㅅ (Siot, 217)
    "\u1107",  # ㅇ (Ieung, 218)
    "\u1108",  # ㅈ (Jieut, 219)
    "\u1109",  # ㅊ (Chieut, 220)
    "\u110a",  # ㅋ (Kieuk, 221)
    "\u110b",  # ㅌ (Tieut, 222)
    "\u110c",  # ㅍ (Pieup, 223)
    "\u110d",  # ㅎ (Hieut, 224)
    # Hangul Vowels (225-245)
    "\u1161",  # ㅏ (A, 225)
    "\u1162",  # ㅐ (Ae, 226)
    "\u1163",  # ㅑ (Ya, 227)
    "\u1164",  # ㅒ (Yae, 228)
    "\u1165",  # ㅓ (Eo, 229)
    "\u1166",  # ㅔ (E, 230)
    "\u1167",  # ㅕ (Yeo, 231)
    "\u1168",  # ㅖ (Ye, 232)
    "\u1169",  # ㅗ (O, 233)
    "\u116a",  # ㅘ (Wa, 234)
    "\u116b",  # ㅙ (Wae, 235)
    "\u116c",  # ㅚ (Oe, 236)
    "\u116d",  # ㅛ (Yo, 237)
    "\u116e",  # ㅜ (U, 238)
    "\u116f",  # ㅝ (Weo, 239)
    "\u1170",  # ㅞ (We, 240)
    "\u1171",  # ㅟ (Wi, 241)
    "\u1172",  # ㅠ (Yu, 242)
    "\u1173",  # ㅡ (Eu, 243)
    "\u1174",  # ㅢ (Ui, 244)
    "\u1175",  # ㅣ (I, 245)
    # Ethiopic Alphabet (246-274)
    "\u12a0",  # አ (Glottal A, 246)
    "\u12a1",  # ኡ (Glottal U, 247)
    "\u12a2",  # ኢ (Glottal I, 248)
    "\u12a3",  # ኣ (Glottal Aa, 249)
    "\u12a4",  # ኤ (Glottal E, 250)
    "\u12a5",  # እ (Glottal Ie, 251)
    "\u12a6",  # ኦ (Glottal O, 252)
    "\u12a7",  # ኧ (Glottal Wa, 253)
    "\u12c8",  # ወ (Wa, 254)
    "\u12c9",  # ዉ (Wu, 255)
    "\u12ca",  # ዊ (Wi, 256)
    "\u12cb",  # ዋ (Waa, 257)
    "\u12cc",  # ዌ (We, 258)
    "\u12cd",  # ው (Wye, 259)
    "\u12ce",  # ዎ (Wo, 260)
    "\u12b0",  # ኰ (Ko, 261)
    "\u12b1",  # ኱ (Ku, 262)
    "\u12b2",  # ኲ (Ki, 263)
    "\u12b3",  # ኳ (Kua, 264)
    "\u12b4",  # ኴ (Ke, 265)
    "\u12b5",  # ኵ (Kwe, 266)
    "\u12b6",  # ኶ (Ko, 267)
    "\u12a0",  # ጐ (Go, 268)
    "\u12a1",  # ጑ (Gu, 269)
    "\u12a2",  # ጒ (Gi, 270)
    "\u12a3",  # መ (Gua, 271)
    "\u12a4",  # ጔ (Ge, 272)
    "\u12a5",  # ጕ (Gwe, 273)
    "\u12a6",  # ጖ (Go, 274)
    # Devanagari Alphabet (275-318)
    "\u0905",  # अ (A, 275)
    "\u0906",  # आ (Aa, 276)
    "\u0907",  # इ (I, 277)
    "\u0908",  # ई (Ii, 278)
    "\u0909",  # उ (U, 279)
    "\u090a",  # ऊ (Uu, 280)
    "\u090b",  # ऋ (R, 281)
    "\u090f",  # ए (E, 282)
    "\u0910",  # ऐ (Ai, 283)
    "\u0913",  # ओ (O, 284)
    "\u0914",  # औ (Au, 285)
    "\u0915",  # क (Ka, 286)
    "\u0916",  # ख (Kha, 287)
    "\u0917",  # ग (Ga, 288)
    "\u0918",  # घ (Gha, 289)
    "\u0919",  # ङ (Nga, 290)
    "\u091a",  # च (Cha, 291)
    "\u091b",  # छ (Chha, 292)
    "\u091c",  # ज (Ja, 293)
    "\u091d",  # झ (Jha, 294)
    "\u091e",  # ञ (Nya, 295)
    "\u091f",  # ट (Ta, 296)
    "\u0920",  # ठ (Tha, 297)
    "\u0921",  # ड (Da, 298)
    "\u0922",  # ढ (Dha, 299)
    "\u0923",  # ण (Na, 300)
    "\u0924",  # त (Ta, 301)
    "\u0925",  # थ (Tha, 302)
    "\u0926",  # द (Da, 303)
    "\u0927",  # ध (Dha, 304)
    "\u0928",  # न (Na, 305)
    "\u092a",  # प (Pa, 306)
    "\u092b",  # फ (Pha, 307)
    "\u092c",  # ब (Ba, 308)
    "\u092d",  # भ (Bha, 309)
    "\u092e",  # म (Ma, 310)
    "\u092f",  # य (Ya, 311)
    "\u0930",  # र (Ra, 312)
    "\u0932",  # ल (La, 313)
    "\u0935",  # व (Va, 314)
    "\u0936",  # श (Sha, 315)
    "\u0937",  # ष (Ssa, 316)
    "\u0938",  # स (Sa, 317)
    "\u0939",  # ह (Ha, 318)
    # Katakana Alphabet (319-364)
    "\u30a2",  # ア (A, 319)
    "\u30a4",  # イ (I, 320)
    "\u30a6",  # ウ (U, 321)
    "\u30a8",  # エ (E, 322)
    "\u30aa",  # オ (O, 323)
    "\u30ab",  # カ (Ka, 324)
    "\u30ad",  # キ (Ki, 325)
    "\u30af",  # ク (Ku, 326)
    "\u30b1",  # ケ (Ke, 327)
    "\u30b3",  # コ (Ko, 328)
    "\u30b5",  # サ (Sa, 329)
    "\u30b7",  # シ (Shi, 330)
    "\u30b9",  # ス (Su, 331)
    "\u30bb",  # セ (Se, 332)
    "\u30bd",  # ソ (So, 333)
    "\u30bf",  # タ (Ta, 334)
    "\u30c1",  # チ (Chi, 335)
    "\u30c4",  # ツ (Tsu, 336)
    "\u30c6",  # テ (Te, 337)
    "\u30c8",  # ト (To, 338)
    "\u30ca",  # ナ (Na, 339)
    "\u30cb",  # ニ (Ni, 340)
    "\u30cc",  # ヌ (Nu, 341)
    "\u30cd",  # ネ (Ne, 342)
    "\u30ce",  # ノ (No, 343)
    "\u30cf",  # ハ (Ha, 344)
    "\u30d2",  # ヒ (Hi, 345)
    "\u30d5",  # フ (Fu, 346)
    "\u30d8",  # ヘ (He, 347)
    "\u30db",  # ホ (Ho, 348)
    "\u30de",  # マ (Ma, 349)
    "\u30df",  # ミ (Mi, 350)
    "\u30e0",  # ム (Mu, 351)
    "\u30e1",  # メ (Me, 352)
    "\u30e2",  # モ (Mo, 353)
    "\u30e4",  # ヤ (Ya, 354)
    "\u30e6",  # ユ (Yu, 355)
    "\u30e8",  # ヨ (Yo, 356)
    "\u30e9",  # ラ (Ra, 357)
    "\u30ea",  # リ (Ri, 358)
    "\u30eb",  # ル (Ru, 359)
    "\u30ec",  # レ (Re, 360)
    "\u30ed",  # ロ (Ro, 361)
    "\u30ef",  # ワ (Wa, 362)
    "\u30f2",  # ヲ (Wo, 363)
    "\u30f3",  # ン (N, 364)
    # Tifinagh Alphabet (365-400)
    "\u2d30",  # ⴰ (Ya, 365)
    "\u2d31",  # ⴱ (Yab, 366)
    "\u2d32",  # ⴲ (Yabh, 367)
    "\u2d33",  # ⴳ (Yag, 368)
    "\u2d34",  # ⴴ (Yagh, 369)
    "\u2d35",  # ⴵ (Yaj, 370)
    "\u2d36",  # ⴶ (Yach, 371)
    "\u2d37",  # ⴷ (Yad, 372)
    "\u2d38",  # ⴸ (Yadh, 373)
    "\u2d39",  # ⴹ (Yadh, emphatic, 374)
    "\u2d3a",  # ⴺ (Yaz, 375)
    "\u2d3b",  # ⴻ (Yazh, 376)
    "\u2d3c",  # ⴼ (Yaf, 377)
    "\u2d3d",  # ⴽ (Yak, 378)
    "\u2d3e",  # ⴾ (Yak, variant, 379)
    "\u2d3f",  # ⴿ (Yaq, 380)
    "\u2d40",  # ⵀ (Yah, 381)
    "\u2d41",  # ⵁ (Yahh, 382)
    "\u2d42",  # ⵂ (Yahl, 383)
    "\u2d43",  # ⵃ (Yahm, 384)
    "\u2d44",  # ⵄ (Yayn, 385)
    "\u2d45",  # ⵅ (Yakh, 386)
    "\u2d46",  # ⵆ (Yakl, 387)
    "\u2d47",  # ⵇ (Yahq, 388)
    "\u2d48",  # ⵈ (Yash, 389)
    "\u2d49",  # ⵉ (Yi, 390)
    "\u2d4a",  # ⵊ (Yij, 391)
    "\u2d4b",  # ⵋ (Yizh, 392)
    "\u2d4c",  # ⵌ (Yink, 393)
    "\u2d4d",  # ⵍ (Yal, 394)
    "\u2d4e",  # ⵎ (Yam, 395)
    "\u2d4f",  # ⵏ (Yan, 396)
    "\u2d50",  # ⵐ (Yang, 397)
    "\u2d51",  # ⵑ (Yany, 398)
    "\u2d52",  # ⵒ (Yap, 399)
    "\u2d53",  # ⵓ (Yu, 400)
    # Sinhala Alphabet (401-444)
    "\u0d85",  # අ (A, 401)
    "\u0d86",  # ආ (Aa, 402)
    "\u0d87",  # ඉ (I, 403)
    "\u0d88",  # ඊ (Ii, 404)
    "\u0d89",  # උ (U, 405)
    "\u0d8a",  # ඌ (Uu, 406)
    "\u0d8b",  # ඍ (R, 407)
    "\u0d8c",  # ඎ (Rr, 408)
    "\u0d8f",  # ඏ (L, 409)
    "\u0d90",  # ඐ (Ll, 410)
    "\u0d91",  # එ (E, 411)
    "\u0d92",  # ඒ (Ee, 412)
    "\u0d93",  # ඓ (Ai, 413)
    "\u0d94",  # ඔ (O, 414)
    "\u0d95",  # ඕ (Oo, 415)
    "\u0d96",  # ඖ (Au, 416)
    "\u0d9a",  # ක (Ka, 417)
    "\u0d9b",  # ඛ (Kha, 418)
    "\u0d9c",  # ග (Ga, 419)
    "\u0d9d",  # ඝ (Gha, 420)
    "\u0d9e",  # ඞ (Nga, 421)
    "\u0d9f",  # ච (Cha, 422)
    "\u0da0",  # ඡ (Chha, 423)
    "\u0da1",  # ජ (Ja, 424)
    "\u0da2",  # ඣ (Jha, 425)
    "\u0da3",  # ඤ (Nya, 426)
    "\u0da4",  # ට (Ta, 427)
    "\u0da5",  # ඥ (Tha, 428)
    "\u0da6",  # ඦ (Da, 429)
    "\u0da7",  # ට (Dha, 430)
    "\u0da8",  # ඨ (Na, 431)
    "\u0daa",  # ඪ (Pa, 432)
    "\u0dab",  # ණ (Pha, 433)
    "\u0dac",  # ඬ (Ba, 434)
    "\u0dad",  # ත (Bha, 435)
    "\u0dae",  # ථ (Ma, 436)
    "\u0daf",  # ද (Ya, 437)
    "\u0db0",  # ධ (Ra, 438)
    "\u0db1",  # ඲ (La, 439)
    "\u0db2",  # ඳ (Va, 440)
    "\u0db3",  # ප (Sha, 441)
    "\u0db4",  # ඵ (Ssa, 442)
    "\u0db5",  # බ (Sa, 443)
    "\u0db6",  # භ (Ha, 444)
]


import os
import json

import pydantic
import pytest

from libtensorshield.types import SS58Address


class Model(pydantic.BaseModel):
    ss58_address: SS58Address


def test_get_model_validate():
    k = os.urandom(32)
    ck = SS58Address.frombytes(k)
    Model.model_validate({'ss58_address': ck})


def test_model_validate_json():
    k = os.urandom(32)
    ck = SS58Address.frombytes(k)
    Model.model_validate_json(json.dumps({'ss58_address': str(ck)}))


def test_serialize():
    adapter: pydantic.TypeAdapter[SS58Address] = pydantic.TypeAdapter(SS58Address)
    k = os.urandom(32)
    ck = SS58Address.frombytes(k)
    adapter.dump_python(ck)


def test_serialize_json():
    adapter: pydantic.TypeAdapter[SS58Address] = pydantic.TypeAdapter(SS58Address)
    k = os.urandom(32)
    ck = SS58Address.frombytes(k)
    adapter.dump_json(ck)


@pytest.mark.parametrize("value", ['foo'])
def test_invalid_input(value: str):
    adapter: pydantic.TypeAdapter[SS58Address] = pydantic.TypeAdapter(SS58Address)
    try:
        adapter.validate_python(value)
        assert False
    except ValueError:
        pass


def test_generate_json_schema():
    adapter: pydantic.TypeAdapter[SS58Address] = pydantic.TypeAdapter(SS58Address)
    adapter.json_schema()


import logging
from typing import Any
from typing import ClassVar

import pydantic

from ._neuron import Neuron
from ._neuroninfo import NeuronInfo


NeuronType = Neuron


class NeuronList(pydantic.BaseModel):
    logger: ClassVar[logging.Logger] = logging.getLogger(__name__)
    block: int = pydantic.Field(
        default=0
    )

    netuid: int = pydantic.Field(
        default=...
    )

    items: list[Neuron] = pydantic.Field(
        default_factory=list
    )

    _index: dict[str, Neuron] = pydantic.PrivateAttr(
        default_factory=dict
    )

    def get(self, hotkey: str) -> Neuron | None:
        return self._index.get(str(hotkey))

    def model_post_init(self, _: Any):
        self._index = {str(x.hotkey): x for x in self.items}

    def update(
        self,
        netuid: int,
        block: int,
        neurons: list[NeuronInfo]
    ):
        adapter: pydantic.TypeAdapter[NeuronType] = pydantic.TypeAdapter(NeuronType)
        current: list[Neuron] = []
        joined: set[NeuronType] = set()
        changed: set[tuple[NeuronType, NeuronType, tuple[str, ...]]] = set()

        for neuroninfo in neurons:
            current.append(
                adapter.validate_python({
                    'hotkey': neuroninfo.hotkey,
                    'coldkey': neuroninfo.coldkey,
                    'uid': neuroninfo.uid,
                    'active': neuroninfo.active,
                    'stake': float(neuroninfo.stake),
                    'total_stake': float(neuroninfo.total_stake),
                    'rank': neuroninfo.rank,
                    'emission': neuroninfo.emission,
                    'inventive': neuroninfo.incentive,
                    'consensus': neuroninfo.consensus,
                    'trust': neuroninfo.trust,
                    'vtrust': neuroninfo.validator_trust,
                    'dividends': neuroninfo.dividends,
                    'last_update': neuroninfo.last_update,
                    'validator_permit': neuroninfo.validator_permit,
                    'host': None if not neuroninfo.axon_info else neuroninfo.axon_info.ip,
                    'port': None if not neuroninfo.axon_info else neuroninfo.axon_info.port,
                })
            )

        joined = {x for x in current if x.hotkey not in self._index} # type: ignore
        index: dict[str, NeuronType] = {x.hotkey: x for x in joined}
        for new in current:
            if new in joined:
                continue
            diff = self._index[new.hotkey].update(new)
            if diff[-1]:
                changed.add(diff)
                self.logger.debug(
                    "Neuron %s/%s changed fields: %s",
                    netuid,
                    new.uid,
                    str.join(', ', diff[-1])
                )
            index[new.hotkey] = self._index[new.hotkey]

        # All neurons that are not in the new index are dropped.
        dropped: set[NeuronType] = {
            n for n in self._index.values() # type: ignore
            if n.hotkey not in index
        }

        # Prevent marking neurons as joined when instantiating an empty
        # NeuronList.
        if not self.items:
            joined = set()

        self._index = index
        self.items = list(sorted(self._index.values(), key=lambda x: x.rank))
        return changed, joined, dropped


from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional

from libtensorshield.utils import decode_account_id
from libtensorshield.utils import u16_normalized_float

# for annotation purposes
from ._axoninfo import AxonInfo
from ._balance import Balance
from ._infobase import InfoBase
if TYPE_CHECKING:
    from ._neuroninfolite import NeuronInfoLite


@dataclass
class NeuronInfo(InfoBase):
    """Represents the metadata of a neuron including keys, UID, stake, rankings, and other attributes.

    Attributes:
        hotkey (str): The hotkey associated with the neuron.
        coldkey (str): The coldkey associated with the neuron.
        uid (int): The unique identifier for the neuron.
        netuid (int): The network unique identifier for the neuron.
        active (int): The active status of the neuron.
        stake (Balance): The balance staked to this neuron.
        stake_dict (dict[str, Balance]): A dictionary mapping coldkey to the amount staked.
        total_stake (Balance): The total amount of stake.
        rank (float): The rank score of the neuron.
        emission (float): The emission rate.
        incentive (float): The incentive value.
        consensus (float): The consensus score.
        trust (float): The trust score.
        validator_trust (float): The validation trust score.
        dividends (float): The dividends value.
        last_update (int): The timestamp of the last update.
        validator_permit (bool): Validator permit status.
        weights (list[tuple[int]]): List of weights associated with the neuron.
        bonds (list[list[int]]): List of bonds associated with the neuron.
        pruning_score (int): The pruning score of the neuron.
        prometheus_info (Optional[PrometheusInfo]): Information related to Prometheus.
        axon_info (Optional[AxonInfo]): Information related to Axon.
        is_null (bool): Indicator if this is a null neuron.
    """

    hotkey: str
    coldkey: str
    uid: int
    netuid: int
    active: int
    stake: "Balance"
    # mapping of coldkey to amount staked to this Neuron
    stake_dict: dict[str, "Balance"]
    total_stake: "Balance"
    rank: float
    emission: float
    incentive: float
    consensus: float
    trust: float
    validator_trust: float
    dividends: float
    last_update: int
    validator_permit: bool
    weights: list[tuple[int, int]]
    bonds: list[list[int]]
    pruning_score: int
    axon_info: Optional["AxonInfo"] = None
    is_null: bool = False

    @classmethod
    def from_weights_bonds_and_neuron_lite(
        cls,
        neuron_lite: "NeuronInfoLite",
        weights_as_dict: dict[int, list[tuple[int, int]]],
        bonds_as_dict: dict[int, list[tuple[int, int]]],
    ) -> "NeuronInfo":
        """
        Creates an instance of NeuronInfo from NeuronInfoLite and dictionaries of weights and bonds.

        Args:
            neuron_lite (NeuronInfoLite): A lite version of the neuron containing basic attributes.
            weights_as_dict (dict[int, list[tuple[int, int]]]): A dictionary where the key is the UID and the value is
                a list of weight tuples associated with the neuron.
            bonds_as_dict (dict[int, list[tuple[int, int]]]): A dictionary where the key is the UID and the value is a
                list of bond tuples associated with the neuron.

        Returns:
            NeuronInfo: An instance of NeuronInfo populated with the provided weights and bonds.
        """
        n_dict = neuron_lite.__dict__
        n_dict["weights"] = weights_as_dict.get(neuron_lite.uid, [])
        n_dict["bonds"] = bonds_as_dict.get(neuron_lite.uid, [])

        return cls(**n_dict)

    @staticmethod
    def get_null_neuron() -> "NeuronInfo":
        """Returns a null NeuronInfo instance."""
        neuron = NeuronInfo(
            uid=0,
            netuid=0,
            active=0,
            stake=Balance.from_rao(0),
            stake_dict={},
            total_stake=Balance.from_rao(0),
            rank=0,
            emission=0,
            incentive=0,
            consensus=0,
            trust=0,
            validator_trust=0,
            dividends=0,
            last_update=0,
            validator_permit=False,
            weights=[],
            bonds=[],
            axon_info=None,
            is_null=True,
            coldkey="000000000000000000000000000000000000000000000000",
            hotkey="000000000000000000000000000000000000000000000000",
            pruning_score=0,
        )
        return neuron

    @classmethod
    def _from_dict(cls, decoded: Any) -> "NeuronInfo":
        """Returns a NeuronInfo object from decoded chain data."""
        stake_dict = Balance.process_stake_data(decoded["stake"])
        total_stake = sum(stake_dict.values()) if stake_dict else Balance(0)
        if not isinstance(total_stake, Balance):
            total_stake = Balance(total_stake)
        coldkey = decode_account_id(decoded["coldkey"])
        hotkey = decode_account_id(decoded["hotkey"])
        return NeuronInfo(
            active=decoded["active"],
            axon_info=AxonInfo.from_dict(
                decoded["axon_info"]
                | {
                    "hotkey": hotkey,
                    "coldkey": coldkey,
                },
            ),
            bonds=[[e[0], e[1]] for e in decoded["bonds"]],
            coldkey=coldkey,
            consensus=u16_normalized_float(decoded["consensus"]),
            dividends=u16_normalized_float(decoded["dividends"]),
            emission=decoded["emission"] / 1e9,
            hotkey=hotkey,
            incentive=u16_normalized_float(decoded["incentive"]),
            is_null=False,
            last_update=decoded["last_update"],
            netuid=decoded["netuid"],
            pruning_score=decoded["pruning_score"],
            rank=u16_normalized_float(decoded["rank"]),
            stake_dict=stake_dict,
            stake=total_stake,
            total_stake=total_stake,
            trust=u16_normalized_float(decoded["trust"]),
            uid=decoded["uid"],
            validator_permit=decoded["validator_permit"],
            validator_trust=u16_normalized_float(decoded["validator_trust"]),
            weights=[(e[0], e[1]) for e in decoded["weights"]],
        )



from ._balance import Balance
from ._neuron import Neuron
from ._neuronlist import NeuronList
from ._neuroninfo import NeuronInfo
from ._neuroninfolite import NeuronInfoLite
from ._ss58address import SS58Address


__all__: list[str] = [
    'Balance',
    'Neuron',
    'NeuronInfo',
    'NeuronInfoLite',
    'NeuronList',
    'SS58Address'
]


from scalecodec.utils.ss58 import ss58_encode


__all__: list[str] = [
    'u16_normalized_float'
]

RAOPERTAO = 1e9

SS58_FORMAT = 42

U16_MAX = 65535

U64_MAX = 18446744073709551615


def u16_normalized_float(x: int) -> float:
    return float(x) / float(U16_MAX)


def decode_account_id(account_id_bytes: bytes | str | tuple[tuple[int, ...]]) -> str:
    """
    Decodes an AccountId from bytes to a Base64 string using SS58 encoding.

    Args:
        account_id_bytes (bytes): The AccountId in bytes that needs to be decoded.

    Returns:
        str: The decoded AccountId as a Base64 string.
    """
    if isinstance(account_id_bytes, tuple):
        account_id_bytes = bytes(account_id_bytes[0])

    if isinstance(account_id_bytes, str):
        account_id_bytes = str.encode(account_id_bytes)

    # Convert the AccountId bytes to a Base64 string
    return ss58_encode(bytes(account_id_bytes).hex(), SS58_FORMAT)


import os

import bittensor

from libtensorshield.types import SS58Address


def test_public_key_is_equal():
    k = os.urandom(32)
    bk = bittensor.Keypair.create_from_seed(k) # type: ignore
    ck = SS58Address(bk.ss58_address) # type: ignore
    assert ck.public_bytes == bk.public_key # type: ignore


def test_validate_signature():
    k = os.urandom(32)
    bk = bittensor.Keypair.create_from_seed(k) # type: ignore
    ck = SS58Address(bk.ss58_address) # type: ignore

    msg = 'Hello world!'
    sig: bytes = bk.sign(msg) # type: ignore

    assert bk.ss58_address == ck # type: ignore
    assert bk.verify(msg, sig) == ck.verify(msg, sig) # type: ignore


def test_validate_signature_hex():
    k = os.urandom(32)
    bk = bittensor.Keypair.create_from_seed(k) # type: ignore
    ck = SS58Address(bk.ss58_address) # type: ignore

    msg = 'Hello world!'
    sig: bytes = '0x' + bytes.hex(bk.sign(msg)) # type: ignore

    assert bk.ss58_address == ck # type: ignore
    assert bk.verify(msg, sig) == ck.verify(msg, sig) # type: ignore


from typing import TypedDict


class FixedPoint(TypedDict):
    """
    Represents a fixed point ``U64F64`` number.
    Where ``bits`` is a U128 representation of the fixed point number.

    This matches the type of the Alpha shares.
    """

    bits: int


"""
This module defines the `AxonInfo` class, a data structure used to represent information about an axon endpoint
in the bittensor network.
"""
import ipaddress
from dataclasses import asdict, dataclass
from typing import Any
from typing import TypeVar
from typing import Union

from async_substrate_interface.utils import json

from ._infobase import InfoBase


T = TypeVar('T', bound='AxonInfo')


@dataclass
class AxonInfo(InfoBase):
    """
    The `AxonInfo` class represents information about an axon endpoint in the bittensor network. This includes
    properties such as IP address, ports, and relevant keys.

    Attributes:
        version (int): The version of the axon endpoint.
        ip (str): The IP address of the axon endpoint.
        port (int): The port number the axon endpoint uses.
        ip_type (int): The type of IP protocol (e.g., IPv4 or IPv6).
        hotkey (str): The hotkey associated with the axon endpoint.
        coldkey (str): The coldkey associated with the axon endpoint.
        protocol (int): The protocol version (default is 4).
        placeholder1 (int): Reserved field (default is 0).
        placeholder2 (int): Reserved field (default is 0).
    """

    version: int
    ip: str
    port: int
    ip_type: int
    hotkey: str
    coldkey: str
    protocol: int = 4
    placeholder1: int = 0
    placeholder2: int = 0

    @property
    def is_serving(self) -> bool:
        """True if the endpoint is serving."""
        return self.ip != "0.0.0.0"

    def ip_str(self) -> str:
        """Return the whole IP as string"""
        if self.protocol != 4:
            raise NotImplementedError
        return f'{self.ip}:{self.port}'

    def __eq__(self, other: object):
        if other is None:
            return False

        if not isinstance(other, AxonInfo):
            return NotImplemented

        if (
            self.version == other.version
            and self.ip == other.ip
            and self.port == other.port
            and self.ip_type == other.ip_type
            and self.coldkey == other.coldkey
            and self.hotkey == other.hotkey
        ):
            return True

        return False

    def __str__(self):
        return f"AxonInfo( {self.ip_str()}, {self.hotkey}, {self.coldkey}, {self.version} )"

    def __repr__(self):
        return self.__str__()

    def to_string(self) -> str:
        """Converts the `AxonInfo` object to a string representation using JSON."""
        try:
            return json.dumps(asdict(self))
        except (TypeError, ValueError):
            return AxonInfo(0, "", 0, 0, "", "").to_string()

    @classmethod
    def _from_dict(cls: type[T], decoded: dict[str, Any]) -> T:
        """Returns a AxonInfo object from decoded chain data."""
        return cls(
            version=decoded["version"],
            ip=str(ipaddress.IPv4Address(int(decoded["ip"]))),
            port=decoded["port"],
            ip_type=decoded["ip_type"],
            placeholder1=decoded["placeholder1"],
            placeholder2=decoded["placeholder2"],
            protocol=decoded["protocol"],
            hotkey=decoded["hotkey"],
            coldkey=decoded["coldkey"],
        )

    @classmethod
    def from_string(cls, json_string: str) -> "AxonInfo":
        """
        Creates an `AxonInfo` object from its string representation using JSON.

        Args:
            json_string (str): The JSON string representation of the AxonInfo object.

        Returns:
            AxonInfo: An instance of AxonInfo created from the JSON string. If decoding fails, returns a default
                `AxonInfo` object with default values.

        Raises:
            json.JSONDecodeError: If there is an error in decoding the JSON string.
            TypeError: If there is a type error when creating the AxonInfo object.
            ValueError: If there is a value error when creating the AxonInfo object.
        """
        try:
            data = json.loads(json_string)
            return cls(**data)
        except (json.JSONDecodeError, TypeError, ValueError):
            return AxonInfo(0, "", 0, 0, "", "")

    @classmethod
    def from_neuron_info(cls, neuron_info: dict[str, Any]) -> "AxonInfo":
        """
        Converts a dictionary to an `AxonInfo` object.

        Args:
            neuron_info (dict): A dictionary containing the neuron information.

        Returns:
            instance (AxonInfo): An instance of AxonInfo created from the dictionary.
        """
        return cls(
            version=neuron_info["axon_info"]["version"],
            ip=str(ipaddress.IPv4Address(int(neuron_info["axon_info"]["ip"]))),
            port=neuron_info["axon_info"]["port"],
            ip_type=neuron_info["axon_info"]["ip_type"],
            hotkey=neuron_info["hotkey"],
            coldkey=neuron_info["coldkey"],
        )

    def to_parameter_dict(
        self,
    ) -> dict[str, Union[int, str]]:
        """Returns a torch tensor or dict of the subnet info, depending on the USE_TORCH flag set."""
        return self.__dict__

    @classmethod
    def from_parameter_dict(
        cls,
        parameter_dict: dict[str, Any]
    ) -> "AxonInfo":
        """Returns an axon_info object from a torch parameter_dict or a parameter dict."""
        return cls(**parameter_dict)



import functools
from typing import Any
from typing import TypeVar

from substrateinterface.utils.ss58 import ss58_decode
from substrateinterface.utils.ss58 import ss58_encode
from substrateinterface.utils.ss58 import is_valid_ss58_address
from substrateinterface import Keypair
from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler


__all__: list[str] = [
    'SS58Address'
]

T = TypeVar('T', bound='SS58Address')


class SS58Address(str):
    __module__: str = 'tensorshield.types'

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(),
            python_schema=core_schema.chain_schema([
                core_schema.is_instance_schema(str),
                core_schema.no_info_plain_validator_function(cls.validate)
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(cls.serialize)
        )

    @staticmethod
    def is_valid(value: str):
        if not isinstance(value, str):
            return False
        return is_valid_ss58_address(value, valid_ss58_format=42)

    @classmethod
    def frombytes(cls, value: bytes):
        return cls(ss58_encode(value, 42))

    @functools.cached_property
    def public_bytes(self):
        return bytes.fromhex(ss58_decode(self, 42))

    @functools.cached_property
    def keypair(self):
        return Keypair(ss58_address=self, ss58_format=42)

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema())

    @classmethod
    def validate(cls: type[T], instance: T | str) -> T:
        if not is_valid_ss58_address(instance, 42):
            raise ValueError("not a valid SS58 address")
        return cls(instance)

    def serialize(self) -> str:
        return self

    def verify(
        self,
        data: str | bytes,
        signature: str | bytes
    ):
        try:
            return self.keypair.verify(data, signature)
        except ValueError:
            return False

    def __repr__(self): # pragma: no cover
        return f'<SS58Address: {str(self)}>'


from dataclasses import dataclass
from typing import Any
from typing import Optional

from libtensorshield.utils import u16_normalized_float
from libtensorshield.utils import decode_account_id
from ._axoninfo import AxonInfo
from ._balance import Balance
from ._infobase import InfoBase


@dataclass
class NeuronInfoLite(InfoBase):
    """
    NeuronInfoLite is a dataclass representing neuron metadata without weights and bonds.

    Attributes:
        hotkey (str): The hotkey string for the neuron.
        coldkey (str): The coldkey string for the neuron.
        uid (int): A unique identifier for the neuron.
        netuid (int): Network unique identifier for the neuron.
        active (int): Indicates whether the neuron is active.
        stake (Balance): The stake amount associated with the neuron.
        stake_dict (dict): Mapping of coldkey to the amount staked to this Neuron.
        total_stake (Balance): Total amount of the stake.
        rank (float): The rank of the neuron.
        emission (float): The emission value of the neuron.
        incentive (float): The incentive value of the neuron.
        consensus (float): The consensus value of the neuron.
        trust (float): Trust value of the neuron.
        validator_trust (float): Validator trust value of the neuron.
        dividends (float): Dividends associated with the neuron.
        last_update (int): Timestamp of the last update.
        validator_permit (bool): Indicates if the neuron has a validator permit.
        prometheus_info (Optional[PrometheusInfo]): Prometheus information associated with the neuron.
        axon_info (Optional[AxonInfo]): Axon information associated with the neuron.
        pruning_score (int): The pruning score of the neuron.
        is_null (bool): Indicates whether the neuron is null.

    Methods:
        get_null_neuron: Returns a NeuronInfoLite object representing a null neuron.
        list_from_vec_u8: Decodes a bytes object into a list of NeuronInfoLite instances.
    """

    hotkey: str
    coldkey: str
    uid: int
    netuid: int
    active: int
    stake: "Balance"
    # mapping of coldkey to amount staked to this Neuron
    stake_dict: dict[str, "Balance"]
    total_stake: "Balance"
    rank: float
    emission: float
    incentive: float
    consensus: float
    trust: float
    validator_trust: float
    dividends: float
    last_update: int
    validator_permit: bool
    axon_info: Optional["AxonInfo"]
    pruning_score: int
    is_null: bool = False

    @staticmethod
    def get_null_neuron() -> "NeuronInfoLite":
        """Returns a null NeuronInfoLite instance."""
        neuron = NeuronInfoLite(
            uid=0,
            netuid=0,
            active=0,
            stake=Balance.from_rao(0),
            stake_dict={},
            total_stake=Balance.from_rao(0),
            rank=0,
            emission=0,
            incentive=0,
            consensus=0,
            trust=0,
            validator_trust=0,
            dividends=0,
            last_update=0,
            validator_permit=False,
            axon_info=None,
            is_null=True,
            coldkey="000000000000000000000000000000000000000000000000",
            hotkey="000000000000000000000000000000000000000000000000",
            pruning_score=0,
        )
        return neuron

    @classmethod
    def _from_dict(cls, decoded: Any) -> "NeuronInfoLite":
        """Returns a NeuronInfoLite object from decoded chain data."""
        coldkey = decode_account_id(decoded["coldkey"])
        hotkey = decode_account_id(decoded["hotkey"])
        stake_dict = Balance.process_stake_data(decoded["stake"])
        stake = sum(stake_dict.values()) if stake_dict else Balance(0)
        if not isinstance(stake, Balance):
            stake = Balance(stake)

        return NeuronInfoLite(
            active=decoded["active"],
            axon_info=AxonInfo.from_dict(
                decoded["axon_info"]
                | {
                    "hotkey": hotkey,
                    "coldkey": coldkey,
                },
            ),
            coldkey=coldkey,
            consensus=u16_normalized_float(decoded["consensus"]),
            dividends=u16_normalized_float(decoded["dividends"]),
            emission=decoded["emission"] / 1e9,
            hotkey=hotkey,
            incentive=u16_normalized_float(decoded["incentive"]),
            last_update=decoded["last_update"],
            netuid=decoded["netuid"],
            pruning_score=decoded["pruning_score"],
            rank=u16_normalized_float(decoded["rank"]),
            stake_dict=stake_dict,
            stake=stake,
            total_stake=stake,
            trust=u16_normalized_float(decoded["trust"]),
            uid=decoded["uid"],
            validator_permit=decoded["validator_permit"],
            validator_trust=u16_normalized_float(decoded["validator_trust"]),
        )



import operator

import pytest

from libtensorshield.types import Balance


@pytest.mark.parametrize("a,b", [
    (1, 1),
    (1.0, 1.0),
    (Balance(1.0), Balance(1.0))
])
def test_eq(a: int | float | Balance, b: int | float | Balance):
    assert Balance(a) == Balance(b)


def test_eq_int():
    assert Balance(1) == 1


def test_ne():
    assert Balance(1) != Balance(2)


def test_gt():
    assert Balance(2) > Balance(1)
    assert Balance(2) > 1.0


def test_lt():
    assert Balance(1) < Balance(2)
    assert Balance(1.0) < Balance(2.0)
    assert Balance(1) < 2

def test_ge():
    assert operator.ge(Balance(1), Balance(1))
    assert operator.ge(Balance(2), Balance(1))
    assert not operator.ge(Balance(1), Balance(2))

def test_le():
    assert operator.le(Balance(1), Balance(1))
    assert operator.le(Balance(1), Balance(2))
    assert not operator.le(Balance(2), Balance(1))


def test_add():
    assert Balance(1) + Balance(1) == Balance(2)
    assert Balance(1.0) + Balance(1.0) == Balance(2.0)
    assert Balance(1) + 1 == Balance(2)


def test_radd():
    assert 1 + Balance(1) == Balance(2)


def test_mul():
    assert Balance(1) * Balance(2) == Balance(2)
    #assert Balance(1.0) * Balance(2.0) == Balance(2.0)
    assert Balance(1) * 2 == Balance(2)
    assert Balance(1.0) * 0.5 == Balance(0.5)


def test_rmul():
    assert 1 * Balance(2) == Balance(2)


def test_sub():
    assert Balance(1) - Balance(1) == Balance(0)
    assert Balance(1.0) - Balance(1.0) == Balance(0)
    assert Balance(1) - 1 == Balance(0)


def test_rsub():
    assert 1 - Balance(1) == Balance(0)


def test_div():
    assert Balance(4) / Balance(2) == Balance(2)
    assert Balance(4.0) / 2 == Balance(2.0)
    assert Balance(4.0) * 0.5 == Balance(2.0)


def test_bool():
    assert bool(Balance(1)) is True
    assert not bool(Balance(0)) is False


def test_abs():
    assert abs(Balance(-1)) == Balance(1)


def test_neg():
    assert -(Balance(1)) == Balance(-1)


def test_rdiv():
    assert 4 / Balance(2) == Balance(2)


def test_eq_none():
    assert not Balance(1) == None


def test_as_tao():
    assert Balance(1000000000).tao == 1


def test_cast_int():
    assert int(Balance(1.0)) == 1000000000


def test_cast_float():
    assert float(Balance(1.0)) == 1.0


def test_from_float():
    assert Balance.from_float(1.0) == Balance(1.0)


def test_from_tao():
    assert Balance.from_tao(1.0) == Balance(1.0)


from dataclasses import dataclass
from typing import Any
from typing import TypeVar

from async_substrate_interface.errors import SubstrateRequestException


T = TypeVar("T", bound="InfoBase")


@dataclass
class InfoBase:
    """Base dataclass for info objects."""

    @classmethod
    def from_dict(cls: type[T], decoded: dict[str, Any]) -> T:
        try:
            return cls._from_dict(decoded)
        except KeyError as e:
            raise SubstrateRequestException(
                f"The {cls} structure is missing {e} from the chain.",
            )

    @classmethod
    def list_from_dicts(cls: type[T], any_list: list[Any]) -> list[T]:
        return [cls.from_dict(any_) for any_ in any_list]

    @classmethod
    def _from_dict(cls: type[T], decoded: dict[str, Any]) -> T:
        return cls(**decoded)



import logging
from typing import ClassVar
from typing import TypeVar

import pydantic
from libcanonical.bases import StateLogger

from ._neuroninfo import NeuronInfo

S = TypeVar('S', bound='Neuron')


class Neuron(pydantic.BaseModel, StateLogger):
    logger: ClassVar[logging.Logger] = logging.getLogger(__name__)
    active: bool = False
    coldkey: str
    consensus: float = 0.0
    dividends: float = 0.0
    emission: float = 0.0
    host: str = '0.0.0.0'
    hotkey: str
    incentive: float = 0.0
    port: int = 0
    rank: float = 0.0
    stake: float = 0.0
    total_stake: float = 0.0
    trust: float = 0.0
    uid: int = -1
    updated: int = 0

    validator_permit: bool = pydantic.Field(
        default=False
    )

    vrust: float = pydantic.Field(
        default=0.0
    )

    @classmethod
    def model_validate_neuroninfo(cls, neuroninfo: NeuronInfo):
        return cls.model_validate({
            'hotkey': neuroninfo.hotkey,
            'coldkey': neuroninfo.coldkey,
            'uid': neuroninfo.uid,
            'active': neuroninfo.active,
            'stake': float(neuroninfo.stake),
            'total_stake': float(neuroninfo.total_stake),
            'rank': neuroninfo.rank,
            'emission': neuroninfo.emission,
            'inventive': neuroninfo.incentive,
            'concensur': neuroninfo.consensus,
            'trust': neuroninfo.trust,
            'vtrust': neuroninfo.validator_trust,
            'dividends': neuroninfo.dividends,
            'last_update': neuroninfo.last_update,
            'validator_permit': neuroninfo.validator_permit,
            'host': None if not neuroninfo.axon_info else neuroninfo.axon_info.ip,
            'port': None if not neuroninfo.axon_info else neuroninfo.axon_info.port
        })

    def is_miner(self) -> bool:
        raise NotImplementedError

    def is_validator(self) -> bool:
        return bool(self.validator_permit) and self.vtrust > 0.0

    def is_registered(self):
        return self.uid >= 0

    def diff(self, neuron: 'Neuron') -> set[str]:
        changed: set[str] = set()
        for attname in self.model_fields.keys():
            if getattr(self, attname) == getattr(neuron, attname):
                continue
            changed.add(attname)
        return changed

    def update(self, neuron: 'Neuron') -> tuple['Neuron', 'Neuron', tuple[str, ...]]:
        adapter: pydantic.TypeAdapter[Neuron] = pydantic.TypeAdapter(Neuron)
        changed = self.diff(neuron)
        old = adapter.validate_python(self.model_dump())
        for k in changed:
            new = getattr(neuron, k)
            setattr(self, k, new)
        return old, neuron, tuple(changed)

    def __hash__(self) -> int:
        return hash(self.hotkey)